import json
from typing import Dict, Sized

import jpype

from neuralogic import is_initialized, initialize
from neuralogic.nn.base import AbstractNeuraLogic
from neuralogic.core.settings import SettingsProxy
from neuralogic.core.enums import Backend


class NeuraLogic(AbstractNeuraLogic):
    def __init__(self, model, dataset_builder, template, settings: SettingsProxy):
        super().__init__(Backend.JAVA, dataset_builder, template, settings)

        if not is_initialized():
            initialize()

        python_strategy = jpype.JClass(
            "cz.cvut.fel.ida.neural.networks.computation.training.strategies.PythonTrainingStrategy"
        )

        self.do_train = True
        self.need_sync = False

        self.neural_model = model
        self.strategy = python_strategy(settings.settings, model)

        optimizer = self.settings.optimizer

        if not optimizer.is_default():
            trainer_field = self.strategy.getClass().getDeclaredField("trainer")
            trainer_field.setAccessible(True)
            trainer = trainer_field.get(self.strategy)

            optimizer_field = trainer.getClass().getSuperclass().getDeclaredField("optimizer")
            optimizer_field.setAccessible(True)
            optimizer_field.set(trainer, optimizer.get())

        self.samples_len = 0
        self.number_format = self.settings.settings_class.superDetailedNumberFormat

        @jpype.JImplements(
            jpype.JClass("cz.cvut.fel.ida.neural.networks.computation.iteration.actions.PythonHookHandler")
        )
        class HookHandler:
            def __init__(self, module: "NeuraLogic"):
                self.module = module

            @jpype.JOverride
            def handleHook(self, hook, value):
                self.module.run_hook(hook, json.loads(value))

        self.hook_handler = HookHandler(self)
        self.reset_parameters()

    def reset_parameters(self):
        self.strategy.resetParameters()

    def train(self):
        self.do_train = True

    def test(self):
        self.do_train = False

    def set_training_samples(self, samples):
        self.samples_len = len(samples)
        self.strategy.setSamples(jpype.java.util.ArrayList(samples))

    def __call__(self, samples=None, train: bool = None, epochs: int = 1):
        self.hooks_set = len(self.hooks) != 0

        if self.hooks_set:
            self.strategy.setHooks(set(self.hooks.keys()), self.hook_handler)

        if train is not None:
            self.do_train = train

        if samples is None:
            results = self.strategy.learnSamples(epochs)
            deserialized_results = json.loads(str(results))

            return deserialized_results, self.samples_len

        if not isinstance(samples, Sized):
            if self.do_train:
                result = self.strategy.learnSample(samples.java_sample)
                return json.loads(str(result)), 1
            return json.loads(str(self.strategy.evaluateSample(samples.java_sample)))

        if self.do_train:
            results = self.strategy.learnSamples(
                jpype.java.util.ArrayList([sample.java_sample for sample in samples]), epochs
            )

            return json.loads(str(results)), len(samples)

        results = self.strategy.evaluateSamples(jpype.java.util.ArrayList([sample.java_sample for sample in samples]))
        return json.loads(str(results))

    def state_dict(self) -> Dict:
        weights = self.neural_model.getAllWeights()
        weights_dict = {}
        weight_names = {}

        for weight in weights:
            if weight.isLearnable:
                value = weight.value

                size = list(value.size())

                if len(size) == 0 or size[0] == 0:
                    weights_dict[weight.index] = value.get(0)
                elif len(size) == 1 or size[0] == 1 or size[1] == 1:
                    weights_dict[weight.index] = list(value.values)
                else:
                    weights_dict[weight.index] = json.loads(str(value.toString()))
                weight_names[weight.index] = weight.name
        return {
            "weights": weights_dict,
            "weight_names": weight_names,
        }

    def load_state_dict(self, state_dict: Dict):
        self.sync_template(state_dict, self.neural_model.getAllWeights())



from typing import Dict, Any


class InitializerNames:
    UNIFORM = "UNIFORM"
    NORMAL = "NORMAL"
    CONSTANT = "CONSTANT"
    LONGTAIL = "LONGTAIL"
    GLOROT = "GLOROT"
    HE = "HE"


class Initializer:
    def is_simple(self) -> bool:
        return True

    def get_settings(self) -> Dict[str, Any]:
        return {"initializer": str(self)}


class Uniform(Initializer):
    r"""Initializes learnable parameters with random uniformly distributed samples from the interval
    ``[-scale / 2, scale / 2]``.

    Parameters
    ----------

    scale : float
        Scale of the distribution interval [-scale / 2, scale / 2]. Default: ``2``

    """
    def __init__(self, scale: float = 2):
        self.scale = scale

    def get_settings(self) -> Dict[str, Any]:
        return {
            "initializer": str(self),
            "initializer_uniform_scale": self.scale,
        }

    def __str__(self):
        return InitializerNames.UNIFORM


class Normal(Initializer):
    r"""Initializes learnable parameters with random samples from a normal (Gaussian) distribution"""
    def __str__(self):
        return InitializerNames.NORMAL


class Constant(Initializer):
    r"""Initializes learnable parameters with the ``value``.

    Parameters
    ----------

    value : float
        Value to fill weights with. Default: ``0.1``

    """
    def __init__(self, value: float = 0.1):
        self.value = value

    def get_settings(self) -> Dict[str, Any]:
        return {
            "initializer": str(self),
            "initializer_const": self.value,
        }

    def __str__(self):
        return InitializerNames.CONSTANT


class Longtail(Initializer):
    """Initializes learnable parameters with random samples from a long tail distribution"""
    def __str__(self):
        return InitializerNames.LONGTAIL


class Glorot(Initializer):
    r"""Initializes learnable parameters with samples from a uniform distribution (from the interval
    ``[-scale / 2, scale / 2]``) using the Glorot method.

    Parameters
    ----------

    scale : float
        Scale of a uniform distribution interval [-scale / 2, scale / 2]. Default: ``2``

    """
    def __init__(self, scale: float = 2):
        self.scale = scale

    def is_simple(self) -> bool:
        return False

    def get_settings(self) -> Dict[str, Any]:
        return {
            "initializer": str(self),
            "initializer_uniform_scale": self.scale,
        }

    def __str__(self):
        return InitializerNames.GLOROT


class He(Initializer):
    r"""Initializes learnable parameters with samples from a uniform distribution (from the interval
    ``[-scale / 2, scale / 2]``) using the He method.

    Parameters
    ----------

    scale : float
        Scale of a uniform distribution interval [-scale / 2, scale / 2]. Default: ``2``

    """
    def __init__(self, scale: float = 2):
        self.scale = scale

    def is_simple(self) -> bool:
        return False

    def get_settings(self) -> Dict[str, Any]:
        return {
            "initializer": str(self),
            "initializer_uniform_scale": self.scale,
        }

    def __str__(self):
        return InitializerNames.HE



import tempfile
from typing import Union, Set, Dict

import jpype

import neuralogic.dataset as datasets
from neuralogic import is_initialized, initialize
from neuralogic.core.builder.builder import Builder
from neuralogic.core.builder.components import BuiltDataset
from neuralogic.core.enums import Backend
from neuralogic.core.constructs.relation import BaseRelation, WeightedRelation
from neuralogic.core.constructs.rule import Rule
from neuralogic.core.constructs.java_objects import JavaFactory
from neuralogic.core.settings import SettingsProxy
from neuralogic.core.sources import Sources


TemplateEntries = Union[BaseRelation, WeightedRelation, Rule]


class DatasetBuilder:
    def __init__(self, parsed_template, java_factory: JavaFactory):
        if not is_initialized():
            initialize()

        self.java_factory = java_factory
        self.parsed_template = parsed_template

        self.grounding_mode = jpype.JClass("cz.cvut.fel.ida.setup.Settings").GroundingMode
        self.logic_sample = jpype.JClass("cz.cvut.fel.ida.logic.constructs.example.LogicSample")

        self.examples_builder = jpype.JClass("cz.cvut.fel.ida.logic.constructs.building.ExamplesBuilder")
        self.queries_builder = jpype.JClass("cz.cvut.fel.ida.logic.constructs.building.QueriesBuilder")

        self.query_counter = 0
        self.examples_counter = 0

        self.hooks: Dict[str, Set] = {}

    def build_queries(self, queries, query_builder):
        logic_samples = []
        one_query_per_example = True

        for query in queries:
            head, conjunction = self.java_factory.get_query(query)
            facts = conjunction.facts

            if head is not None:
                id = head.literal.toString()
                if head.getValue() is None or not isinstance(query.head.weight, (float, int)):
                    logic_samples.extend(
                        [self.logic_sample(f.getValue(), query_builder.createQueryAtom(id, f), True) for f in facts]
                    )
                else:
                    importance = head.getValue().value
                    logic_samples.extend(
                        [
                            self.logic_sample(f.getValue(), query_builder.createQueryAtom(id, importance, f), True)
                            for f in facts
                        ]
                    )
            else:
                id = str(self.query_counter)
                if len(facts) > 1:
                    one_query_per_example = False

                logic_samples.extend(
                    [self.logic_sample(f.getValue(), query_builder.createQueryAtom(id, f), True) for f in facts]
                )
            self.query_counter += 1
        return logic_samples, one_query_per_example

    def build_examples(self, examples, examples_builder):
        logic_samples = []
        one = jpype.JClass("cz.cvut.fel.ida.algebra.values.ScalarValue")(1.0)
        examples_queries = False

        for example in examples:
            label, lifted_example = self.java_factory.get_lifted_example(example)
            example_query = False

            value = one
            label_fact = None if label is None else label.facts
            label_size = 0 if label is None else label_fact.size()

            if label is None or label_size == 0:
                query_atom = examples_builder.createQueryAtom(str(self.examples_counter), None, lifted_example)
            elif label_size == 1:
                example_query = True
                if label_fact.get(0).getValue() is None:
                    literal_string = label_fact.get(0).literal.toString()
                    query_atom = examples_builder.createQueryAtom(literal_string, label_fact.get(0), lifted_example)
                else:
                    value = label_fact.get(0).getValue()
                    query_atom = examples_builder.createQueryAtom(
                        str(self.examples_counter), label_fact.get(0), lifted_example
                    )
            else:
                raise NotImplementedError

            if not example_query and examples_queries:
                raise Exception("Inconsistent examples! Some examples have queries and some do not")

            examples_queries = example_query
            logic_samples.append(self.logic_sample(value, query_atom))
            self.examples_counter += 1
        return logic_samples, examples_queries

    def build_dataset(
        self, dataset: datasets.BaseDataset, backend: Backend, settings: SettingsProxy, file_mode: bool = False
    ) -> BuiltDataset:
        """Builds the dataset (does grounding and neuralization) for this template instance and the backend

        :param dataset:
        :param backend:
        :param settings:
        :param file_mode:
        :return:
        """
        if isinstance(dataset, datasets.TensorDataset):
            if file_mode:
                with tempfile.NamedTemporaryFile(mode="w", suffix=".txt") as q_tf, tempfile.NamedTemporaryFile(
                    mode="w", suffix=".txt"
                ) as e_tf:
                    dataset.dump(q_tf, e_tf)

                    q_tf.flush()
                    e_tf.flush()

                    return self.build_dataset(datasets.FileDataset(e_tf.name, q_tf.name), backend, settings, False)

        if isinstance(dataset, datasets.ConvertableDataset):
            return self.build_dataset(dataset.to_dataset(), backend, settings, False)

        if isinstance(dataset, datasets.Dataset):
            self.examples_counter = 0
            self.query_counter = 0

            weight_factory = self.java_factory.weight_factory

            examples_builder = self.examples_builder(settings.settings)
            query_builder = self.queries_builder(settings.settings)
            query_builder.setFactoriesFrom(examples_builder)

            settings.settings.groundingMode = self.grounding_mode.INDEPENDENT

            examples = dataset.examples
            queries = dataset.queries

            if len(examples) == 1:
                settings.settings.groundingMode = self.grounding_mode.GLOBAL
            settings.settings.infer()
            settings._setup_random_generator()

            self.java_factory.weight_factory = self.java_factory.get_new_weight_factory()
            examples, example_queries = self.build_examples(examples, examples_builder)

            self.java_factory.weight_factory = self.java_factory.get_new_weight_factory()
            queries, one_query_per_example = self.build_queries(queries, query_builder)

            logic_samples = DatasetBuilder.merge_queries_with_examples(
                queries, examples, one_query_per_example, example_queries
            )
            logic_samples = jpype.java.util.ArrayList(logic_samples).stream()

            samples = Builder(settings).from_logic_samples(self.parsed_template, logic_samples, backend)

            self.java_factory.weight_factory = weight_factory
        elif isinstance(dataset, datasets.FileDataset):
            args = ["-t", dataset.examples_file or dataset.queries_file]
            if dataset.queries_file is not None:
                args.extend(["-q", dataset.queries_file])
            if dataset.examples_file is not None:
                args.extend(["-e", dataset.examples_file])
            sources = Sources.from_args(args, settings)
            samples = Builder(settings).from_sources(self.parsed_template, sources, backend)
        else:
            raise NotImplementedError

        return BuiltDataset(samples)

    @staticmethod
    def merge_queries_with_examples(queries, examples, one_query_per_example, example_queries=True):
        if len(examples) == 0:
            return queries

        if len(queries) == 0:
            if not example_queries:
                raise Exception("No queries provided! The query list is empty and examples do not contain queries")
            return examples

        # One large example for one or more queries
        if len(examples) == 1:
            logic_samples = []
            for query in queries:
                example = examples[0] if query.query.evidence is None else query
                query_object = query if query.isQueryOnly else examples[0]
                query_object.query.evidence = example.query.evidence
                logic_samples.append(query)
            return logic_samples

        # One example per query
        if one_query_per_example:
            if len(examples) != len(queries):
                raise Exception(
                    f"The size of examples {len(examples)} doesn't match the size of queries {len(queries)}"
                )

            logic_samples = []

            for query, example in zip(queries, examples):
                example_object = example if query.query.evidence is None else query
                query_object = query if query.isQueryOnly else example
                query_object.query.evidence = example_object.query.evidence
                logic_samples.append(query)
            return logic_samples

        logic_samples = []
        example_map = {e.getId(): e for e in examples}

        # Multiple queries per example
        for query in queries:
            query_id = query.getId()

            example_object = example_map[query_id] if query.query.evidence is None else query
            query_object = query if query.isQueryOnly else example_map[query_id]
            query_object.query.evidence = example_object.query.evidence
            logic_samples.append(query)
        return logic_samples



from typing import List, Optional

from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.constructs.function import Transformation, Aggregation
from neuralogic.core.constructs.factories import R, V
from neuralogic.nn.module.module import Module


class RGCNConv(Module):
    r"""
    Relational Graph Convolutional layer from
    `Modeling Relational Data with Graph Convolutional Networks <https://arxiv.org/abs/1703.06103>`_.
    Which can be expressed as:

    .. math::
        \mathbf{x}^{\prime}_i = act(\mathbf{W_0} \cdot \mathbf{x}_i + \sum_{r \in \mathcal{R}}
        {agg}_{j \in \mathcal{N}_r(i)}(\mathbf{W_r} \cdot \mathbf{x}_j))

    Where *act* is an activation function, *agg* aggregation function (by default average), :math:`W_0` is a
    learnable root parameter and :math:`W_r` is a learnable parameter for each relation.

    The first part of the equation that is ":math:`\mathbf{W_0} \cdot \mathbf{x}_i`" can be expressed
    in the logic form as:

    .. code-block:: logtalk

        R.<output_name>(V.I) <= R.<feature_name>(V.I)[<W0>]

    Another part of the equation that is ":math:`{agg}_{j \in \mathcal{N}_r(i)}(\mathbf{W_r} \cdot \mathbf{x}_j)`"
    can be expressed as:

    .. code-block:: logtalk

        R.<output_name>(V.I) <= (R.<feature_name>(V.J)[<Wr>], R.<edge_name>(V.J, relation, V.I))

    where "relation" is a constant name, or as:

    .. code-block:: logtalk

        R.<output_name>(V.I) <= (R.<feature_name>(V.J)[<Wr>], R.<relation>(V.J, V.I))

    The outer summation, together with summing it with the first part, is handled by aggregation of all rules with the
    same head (and substitution).

    Examples
    --------

    The whole computation of this module
    (parametrized as :code:`RGCNConv(1, 2, "h1", "h0", "_edge", ["sibling", "parent"])`) is as follows:

    .. code:: logtalk

        metadata = Metadata(activation=Transformation.IDENTITY, aggregation=Aggregation.AVG)

        (R.h1(V.I) <= R.h0(V.I)[2, 1]) | metadata
        (R.h1(V.I) <= (R.h0(V.J)[2, 1], R._edge(V.J, sibling, V.I))) | metadata
        (R.h1(V.I) <= (R.h0(V.J)[2, 1], R._edge(V.J, parent, V.I))) | metadata
        R.h1 / 1 [Transformation.IDENTITY]

    Module parametrized as :code:`RGCNConv(1, 2, "h1", "h0", None, ["sibling", "parent"])` translates into:

    .. code:: logtalk

        metadata = Metadata(activation=Transformation.IDENTITY, aggregation=Aggregation.AVG)

        (R.h1(V.I) <= R.h0(V.I)[2, 1]) | metadata
        (R.h1(V.I) <= (R.h0(V.J)[2, 1], R.sibling(V.J, V.I))) | metadata
        (R.h1(V.I) <= (R.h0(V.J)[2, 1], R.parent(V.J, V.I))) | metadata
        R.h1 / 1 [Transformation.IDENTITY]

    Parameters
    ----------

    in_channels : int
        Input feature size.
    out_channels : int
        Output feature size.
    output_name : str
        Output (head) predicate name of the module.
    feature_name : str
        Feature predicate name to get features from.
    edge_name : Optional[str]
        Edge predicate name to use for neighborhood relations. When :code:`None`, elements from :code:`relations`
        are used instead.
    relations : List[str]
        List of relations' names
    activation : Transformation
        Activation function of the output.
        Default: ``Transformation.IDENTITY``
    aggregation : Aggregation
        Aggregation function of nodes' neighbors.
        Default: ``Aggregation.SUM``

    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        output_name: str,
        feature_name: str,
        edge_name: Optional[str],
        relations: List[str],
        activation: Transformation = Transformation.IDENTITY,
        aggregation: Aggregation = Aggregation.AVG,
    ):
        self.output_name = output_name
        self.feature_name = feature_name
        self.edge_name = edge_name

        self.relations = relations

        self.in_channels = in_channels
        self.out_channels = out_channels

        self.aggregation = aggregation
        self.activation = activation

    def __call__(self):
        head = R.get(self.output_name)(V.I)
        metadata = Metadata(transformation=Transformation.IDENTITY, aggregation=self.aggregation)
        feature = R.get(self.feature_name)(V.J)[self.out_channels, self.in_channels]

        if self.edge_name is not None:
            relation_rules = [
                ((head <= (feature, R.get(self.edge_name)(V.J, relation, V.I))) | metadata)
                for relation in self.relations
            ]
        else:
            relation_rules = [
                ((head <= (feature, R.get(relation)(V.J, V.I))) | metadata) for relation in self.relations
            ]

        return [
            (head <= R.get(self.feature_name)(V.I)[self.out_channels, self.in_channels]) | metadata,
            *relation_rules,
            R.get(self.output_name) / 1 | Metadata(transformation=self.activation),
        ]



from typing import Dict
from neuralogic.core.constructs.predicate import Predicate
from neuralogic.core.constructs import relation


class AtomFactory:
    class Predicate:
        def __init__(self, hidden=False, special=False):
            self.is_hidden = hidden
            self.is_special = special

        @property
        def special(self) -> "AtomFactory.Predicate":
            return AtomFactory.Predicate(self.is_hidden, True)

        @property
        def hidden(self) -> "AtomFactory.Predicate":
            return AtomFactory.Predicate(True, self.is_special)

        @staticmethod
        def get_predicate(name, arity, hidden, special) -> Predicate:
            return Predicate(name, arity, hidden, special)

        def __getattr__(self, item):
            return relation.BaseRelation(Predicate(item, 0, self.is_hidden, self.is_special))

    def __init__(self):
        self.instances: Dict[str, Dict[int, relation.BaseRelation]] = {}

        self.special = AtomFactory.Predicate(special=True)
        self.hidden = AtomFactory.Predicate(hidden=True)

    def get(self, name: str) -> relation.BaseRelation:
        return relation.BaseRelation(Predicate(name, 0, False, False))

    def __getattr__(self, item) -> relation.BaseRelation:
        return relation.BaseRelation(Predicate(item, 0, False, False))


class VariableFactory:
    def __getattr__(self, item) -> str:
        return item.upper()


class ConstantFactory:
    def __getattr__(self, item) -> str:
        return item.lower()


Var = VariableFactory()
Relation = AtomFactory()
Constant = ConstantFactory()

V = Var
C = Constant
R = Relation



class ErrorFunctionNames:
    MSE = "SQUARED_DIFF"
    CROSSENTROPY = "CROSSENTROPY"
    SOFTENTROPY = "SOFTENTROPY"


class ErrorFunction:
    pass


class MSE(ErrorFunction):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return ErrorFunctionNames.MSE


class CrossEntropy(ErrorFunction):
    def __init__(self, with_logits: bool = True):
        super().__init__()
        self.with_logits = with_logits

    def __str__(self):
        return ErrorFunctionNames.SOFTENTROPY if self.with_logits else ErrorFunctionNames.CROSSENTROPY


class SoftEntropy(ErrorFunction):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return ErrorFunctionNames.SOFTENTROPY



from neuralogic.inference.evaluation_inference_engine import EvaluationInferenceEngine
from neuralogic.inference.inference_engine import InferenceEngine

__all__ = ["InferenceEngine", "EvaluationInferenceEngine"]



from typing import Tuple

import jpype

from neuralogic.optim.optimizer import Optimizer


class Adam(Optimizer):
    def __init__(self, lr: float = 0.001, betas: Tuple[float, float] = (0.9, 0.999), eps: float = 1e-08):
        self._lr = lr
        self._betas = betas
        self._eps = eps

    @property
    def lr(self) -> float:
        return self._lr

    @property
    def betas(self) -> Tuple[float, float]:
        return self._betas

    @property
    def eps(self) -> float:
        return self._eps

    def get(self):
        adam_class = jpype.JClass("cz.cvut.fel.ida.neural.networks.computation.training.optimizers.Adam")
        lr = jpype.JClass("cz.cvut.fel.ida.algebra.values.ScalarValue")(self._lr)

        return adam_class(lr, self._betas[0], self._betas[1], self._eps)

    def is_default(self) -> bool:
        return self._betas == (0.9, 0.999) and self._eps == 1e-08

    def __str__(self) -> str:
        return f"Adam(lr={self.lr}, betas={self.betas}, eps={self.eps})"



from typing import Dict, Optional, Union, Callable, List

from neuralogic.core.settings import Settings
from neuralogic.core.builder import DatasetBuilder
from neuralogic.core import Template, Backend, BuiltDataset, SettingsProxy
from neuralogic.dataset.base import BaseDataset

from neuralogic.utils.visualize import draw_model


class AbstractNeuraLogic:
    def __init__(self, backend: Backend, dataset_builder: DatasetBuilder, template: Template, settings: SettingsProxy):
        self.need_sync = True

        self.source_template = [rule for rule in template.template]
        self.template = dataset_builder.parsed_template
        self.dataset_builder = dataset_builder

        self.backend = backend
        self.settings = settings

        self.hooks_set = False
        self.hooks: Dict[str, List[Callable]] = {}

    def __call__(self, sample):
        raise NotImplementedError

    def build_dataset(self, dataset: BaseDataset, file_mode: bool = False):
        return self.dataset_builder.build_dataset(dataset, self.backend, self.settings, file_mode)

    def set_hooks(self, hooks):
        self.hooks_set = len(hooks) != 0
        self.hooks = hooks

    def run_hook(self, hook: str, value):
        for callback in self.hooks[hook]:
            callback(value)

    def sync_template(self, state_dict: Optional[Dict] = None, weights=None):
        state_dict = self.state_dict() if state_dict is None else state_dict
        weights = self.template.getAllWeights() if weights is None else weights
        weight_dict = state_dict["weights"]

        for weight in weights:
            if not weight.isLearnable:
                continue
            weight_value = weight.value

            index = weight.index
            value = weight_dict[index]

            if isinstance(value, (float, int)):
                weight_value.set(0, float(value))
                continue

            if isinstance(value[0], (float, int)):
                for i, val in enumerate(value):
                    weight_value.set(i, float(val))
                continue

            cols = len(value[0])

            for i, values in enumerate(value):
                for j, val in enumerate(values):
                    weight_value.set(i * cols + j, float(val))

    def parameters(self) -> Dict:
        return self.state_dict()

    def state_dict(self) -> Dict:
        raise NotImplementedError

    def load_state_dict(self, state_dict: Dict):
        raise NotImplementedError

    def draw(
        self,
        filename: Optional[str] = None,
        draw_ipython=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        return draw_model(self, filename, draw_ipython, img_type, value_detail, graphviz_path, *args, **kwargs)


class AbstractEvaluator:
    def __init__(self, backend: Backend, template: Template, settings: Settings):
        self.settings = settings.create_proxy()
        self.backend = backend
        self.dataset: Optional[BuiltDataset] = None

        self.neuralogic_model = template.build(settings, backend)
        self.neuralogic_model.set_hooks(template.hooks)

    def set_dataset(self, dataset: Union[BaseDataset, BuiltDataset]):
        self.dataset = self.build_dataset(dataset)

    def build_dataset(self, dataset: Union[BaseDataset, BuiltDataset], file_mode: bool = False):
        if isinstance(dataset, BaseDataset):
            return self.neuralogic_model.build_dataset(dataset, file_mode)
        return dataset

    @property
    def model(self) -> AbstractNeuraLogic:
        return self.neuralogic_model

    def train(self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True):
        pass

    def test(self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True):
        pass

    def parameters(self) -> Dict:
        return self.state_dict()

    def state_dict(self) -> Dict:
        pass

    def load_state_dict(self, state_dict: Dict):
        pass

    def reset_parameters(self):
        self.neuralogic_model.reset_parameters()

    def draw(
        self,
        filename: Optional[str] = None,
        draw_ipython=True,
        img_type="png",
        value_detail: int = 0,
        graphviz_path: Optional[str] = None,
        *args,
        **kwargs,
    ):
        return self.neuralogic_model.draw(
            filename, draw_ipython, img_type, value_detail, graphviz_path, *args, **kwargs
        )



from typing import List, Optional, Dict, Union, Tuple
import torch
import numpy as np

from neuralogic.core.builder import Sample, Weight, Neuron
from neuralogic.core.settings import SettingsProxy
from neuralogic.core.enums import Backend
from neuralogic.nn.base import AbstractNeuraLogic
from neuralogic.nn.init import InitializerNames


def longtail(tensor: torch.Tensor, _: SettingsProxy):
    init = np.random.rand(*tensor.size())

    x0, x1, power = 0.0, 10.0, 50.0
    x0_power = np.power(x0, power + 1)

    init = x1 - np.power((np.power(x1, power + 1) - x0_power) * init + x0_power, 1 / (power + 1))
    tensor.fill_(init)


class NeuraLogic(AbstractNeuraLogic):
    activations = {
        "Sigmoid": torch.sigmoid,
        "ReLu": torch.relu,
        "Tanh": torch.tanh,
        "Average": torch.mean,
        "Maximum": torch.max,
        "Sum": torch.sum,
        "Minimum": torch.min,
    }

    initializers = {
        InitializerNames.NORMAL: lambda tensor, _: torch.nn.init.normal_(tensor),
        InitializerNames.UNIFORM: lambda tensor, settings: torch.nn.init.uniform_(
            tensor, -settings.initializer_uniform_scale / 2, settings.initializer_uniform_scale / 2
        ),
        InitializerNames.CONSTANT: lambda tensor, settings: torch.nn.init.constant_(tensor, settings.initializer_const),
        InitializerNames.LONGTAIL: longtail,
        InitializerNames.GLOROT: lambda tensor, _: torch.nn.init.xavier_uniform_(tensor),
        InitializerNames.HE: lambda tensor, _: torch.nn.init.kaiming_uniform_(tensor),
    }

    def __init__(self, model: List[Weight], dataset_builder, template, settings: Optional[SettingsProxy] = None):
        super().__init__(Backend.TORCH, dataset_builder, template, settings)

        self.model: Optional[torch.nn.ParameterList] = None
        self.weights_meta = model
        self.weights: List[torch.nn.Parameter] = []

        self.reset_parameters()

    def reset_parameters(self):
        initializer = str(self.settings.initializer)

        if initializer not in NeuraLogic.initializers:
            raise NotImplementedError

        weight_initializer = NeuraLogic.initializers[initializer]
        self.weights = []

        for weight in self.weights_meta:
            if weight.fixed:
                parameter = torch.nn.Parameter(torch.tensor(weight.value, dtype=torch.float64), requires_grad=False)
            else:
                parameter = torch.nn.Parameter(
                    torch.empty(size=weight.dimensions, requires_grad=True, dtype=torch.float64)
                )
                weight_initializer(parameter.data, self.settings)
            self.weights.append(parameter)
        self.model = torch.nn.ParameterList(self.weights)

    def build_sample(self, sample: Sample):
        self.hooks_set = len(self.hooks) != 0

        torch_neurons: List[Optional[torch.Tensor]] = [None] * len(sample.neurons)

        for neuron in sample.neurons:
            torch_neurons[neuron.index] = self.to_torch_expression(neuron, torch_neurons, self.model)
        return torch_neurons[sample.neurons[-1].index]

    def __call__(self, sample: Sample):
        return self.build_sample(sample)

    def state_dict(self) -> Dict:
        weights = {}

        for meta, weight in zip(self.weights_meta, self.weights):
            if not meta.fixed and meta.index >= 0:
                weights[meta.index] = weight if isinstance(weight, (int, float)) else weight.data
        return {"weights": weights}

    def load_state_dict(self, state_dict: Dict):
        weight_dict = state_dict["weights"]

        for i, (meta, weight) in enumerate(zip(self.weights_meta, self.weights)):
            if meta.index < 0 or meta.fixed:
                continue

            if isinstance(weight, (int, float)):
                weight.data = torch.tensor([weight_dict[meta.index]], dtype=torch.float64)
            else:
                weight.data = torch.tensor(weight_dict[meta.index], dtype=torch.float64)

    @staticmethod
    def to_tensor_value(value) -> torch.Tensor:
        if isinstance(value, (float, int)):
            return torch.tensor([value], requires_grad=False, dtype=torch.float64)
        return torch.tensor(value, requires_grad=False, dtype=torch.float64)

    def to_torch_expression(
        self, neuron: Neuron, neurons: List[torch.Tensor], weights: torch.nn.ParameterList
    ) -> torch.Tensor:
        if neuron.inputs:
            out_scalar, out_vector, out_matrix = self.process_neuron_inputs(neuron, neurons, weights)

            if neuron.activation:
                if not neuron.pooling:
                    out = sum(out_scalar) + sum(out_vector) + sum(out_matrix)
                else:
                    if out_matrix:
                        shape = out_matrix[0].shape
                        dummy_tensor = torch.zeros((shape), requires_grad=False)

                        out = torch.stack(
                            [
                                *out_matrix,
                                *(vector + dummy_tensor for vector in out_vector),
                                *(scalar + dummy_tensor for scalar in out_scalar),
                            ]
                        )
                    elif out_vector:
                        shape = out_vector[0].shape
                        dummy_tensor = torch.zeros((shape), requires_grad=False)
                        out = torch.stack([*out_vector, *(scalar + dummy_tensor for scalar in out_scalar)])
                    else:
                        out = torch.stack(out_scalar)
                if neuron.activation != "Identity":
                    if neuron.pooling:
                        out = NeuraLogic.activations[neuron.activation](out, dim=0)
                    else:
                        out = NeuraLogic.activations[neuron.activation](out)
            else:
                out = sum(out_scalar) + sum(out_vector) + sum(out_matrix)
        else:
            out = NeuraLogic.to_tensor_value(neuron.value)

            if neuron.activation and neuron.activation != "Identity":
                if neuron.pooling:
                    out = NeuraLogic.activations[neuron.activation](out, dim=0)
                else:
                    out = NeuraLogic.activations[neuron.activation](out)

        if self.hooks_set and neuron.hook_name is not None and neuron.hook_name in self.hooks:
            self.run_hook(neuron.hook_name, out.value())

        return out

    def process_neuron_inputs(
        self, neuron: Neuron, neurons: List[torch.Tensor], weights: torch.nn.ParameterList
    ) -> Tuple[List[Union[torch.Tensor, int, float]], List[torch.Tensor], List[torch.Tensor]]:
        out_matrix = []
        out_vector = []
        out_scalar = []

        if neuron.weights:
            for w, i in zip(neuron.weights, neuron.inputs):
                neuron = neurons[i]
                weight = weights[w]

                neuron_size = neuron.size()
                weight_size = weight.size()

                if not neuron_size or not weight_size or neuron_size == (1,) or weight_size == (1,):
                    result = torch.multiply(weight, neuron)
                else:
                    result = torch.matmul(weight, neuron)

                res_size = result.size()
                if not res_size or res_size == (1,):
                    out_scalar.append(result)
                elif len(res_size) == 1:
                    out_vector.append(result)
                else:
                    out_matrix.append(result)
        else:
            for i in neuron.inputs:
                result = neurons[i]

                if isinstance(result, (int, float)):
                    out_scalar.append(result)
                else:
                    res_size = result.size()
                    if not res_size or res_size == (1,):
                        out_scalar.append(result)
                    elif len(res_size) == 1:
                        out_vector.append(result)
                    else:
                        out_matrix.append(result)
        return out_scalar, out_vector, out_matrix



from typing import Optional, Dict, Union

import torch
import torch.nn.functional

from neuralogic.nn.loss import ErrorFunctionNames
from neuralogic.dataset.base import BaseDataset
from neuralogic.nn.base import AbstractEvaluator

from neuralogic.core import Template, BuiltDataset
from neuralogic.core.settings import Settings
from neuralogic.core.enums import Backend


class TorchEvaluator(AbstractEvaluator):
    trainers = {
        "SGD": lambda param, optimizer: torch.optim.SGD(param, lr=optimizer.lr),
        "Adam": lambda param, optimizer: torch.optim.Adam(
            param, lr=optimizer.lr, betas=optimizer.betas, eps=optimizer.eps
        ),
    }

    error_functions = {
        str(ErrorFunctionNames.MSE): torch.nn.MSELoss(),
        str(ErrorFunctionNames.SOFTENTROPY): torch.nn.CrossEntropyLoss(),
        # ErrorFunction.ABS_DIFF: lambda out, target: dy.abs(out - target),
        # ErrorFunction.CROSSENTROPY: lambda out, target: pass
    }

    def __init__(
        self,
        template: Template,
        settings: Settings,
    ):
        super().__init__(Backend.TORCH, template, settings)

    def train(self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        epochs = self.settings.epochs
        error_function = str(self.settings.error_function)
        optimizer = self.settings.optimizer

        if optimizer.name() not in TorchEvaluator.trainers:
            raise NotImplementedError
        if error_function not in TorchEvaluator.error_functions:
            raise NotImplementedError

        trainer = TorchEvaluator.trainers[optimizer.name()](self.neuralogic_model.model, optimizer)
        error_function = TorchEvaluator.error_functions[error_function]

        def _train():
            for _ in range(epochs):
                seen_instances = 0
                total_loss = 0
                for sample in dataset.samples:
                    trainer.zero_grad(set_to_none=True)

                    if isinstance(sample.target, (int, float)):
                        label = torch.tensor([sample.target], dtype=torch.float64, requires_grad=False)
                    else:
                        label = torch.tensor(sample.target, dtype=torch.float64, requires_grad=False)

                    graph_output = self.neuralogic_model(sample)
                    loss = error_function(graph_output, label)

                    try:
                        loss.backward()
                    except RuntimeError:
                        pass
                    trainer.step()

                    total_loss += loss.item()
                    seen_instances += 1
                yield total_loss, seen_instances

        if generator:
            return _train()

        stats = 0, 0
        for stats in _train():
            pass
        return stats

    def test(self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        def _test():
            with torch.no_grad():
                for sample in dataset.samples:
                    graph_output = self.neuralogic_model(sample)

                    if graph_output.size() == (1,):
                        yield graph_output[0].item()
                    elif not graph_output.size():
                        yield graph_output.item()
                    else:
                        yield graph_output

        if generator:
            return _test()
        return list(_test())

    def state_dict(self) -> Dict:
        return self.neuralogic_model.state_dict()

    def load_state_dict(self, state_dict: Dict):
        self.neuralogic_model.load_state_dict(state_dict)



from typing import List, Optional, Dict
import dynet as dy
import numpy as np

from neuralogic.core.builder import Sample, Weight, Neuron
from neuralogic.core.settings import SettingsProxy
from neuralogic.core.enums import Backend
from neuralogic.nn.init import InitializerNames
from neuralogic.nn.base import AbstractNeuraLogic


def longtail(model, weight: Weight, _: SettingsProxy):
    init = np.random.rand(*weight.dimensions)

    x0, x1, power = 0.0, 10.0, 50.0
    x0_power = np.power(x0, power + 1)

    init = x1 - np.power((np.power(x1, power + 1) - x0_power) * init + x0_power, 1 / (power + 1))

    return model.add_parameters(weight.dimensions, init=init)


class NeuraLogic(AbstractNeuraLogic):
    activations = {
        "Sigmoid": dy.logistic,
        "Average": dy.average,
        "Maximum": dy.emax,
        "ReLu": dy.rectify,
        "Tanh": dy.tanh,
    }

    initializers = {
        InitializerNames.HE: lambda model, weight, _: model.add_parameters(weight.dimensions, init="he"),
        InitializerNames.GLOROT: lambda model, weight, _: model.add_parameters(weight.dimensions, init="glorot"),
        InitializerNames.NORMAL: lambda model, weight, _: model.add_parameters(weight.dimensions, init="normal"),
        InitializerNames.UNIFORM: lambda model, weight, settings: model.add_parameters(
            weight.dimensions, init="uniform", scale=settings.initializer_uniform_scale
        ),
        InitializerNames.CONSTANT: lambda model, weight, settings: model.add_parameters(
            weight.dimensions, init=settings.initializer_const
        ),
        InitializerNames.LONGTAIL: longtail,
    }

    def __init__(self, model: List[Weight], dataset_builder, template, settings: Optional[SettingsProxy] = None):
        super().__init__(Backend.DYNET, dataset_builder, template, settings)

        self.model: Optional[dy.ParameterCollection] = None
        self.weights_meta = model
        self.weights: List[dy.Parameters] = []

        self.reset_parameters()

    def reset_parameters(self):
        initializer = str(self.settings.initializer)

        if initializer not in NeuraLogic.initializers:
            raise NotImplementedError

        weight_initializer = NeuraLogic.initializers[initializer]

        self.model = dy.ParameterCollection()
        self.weights = [
            self.model.add_parameters(weight.dimensions, init=np.array(weight.value))
            if weight.fixed
            else weight_initializer(self.model, weight, self.settings)
            for weight in self.weights_meta
        ]

    def build_sample(self, sample: Sample) -> dy.Expression:
        self.hooks_set = len(self.hooks) != 0

        dynet_neurons: List[Optional[dy.Expression]] = [None] * len(sample.neurons)

        for neuron in sample.neurons:
            dynet_neurons[neuron.index] = self.to_dynet_expression(neuron, dynet_neurons, self.weights)
        return dynet_neurons[sample.neurons[-1].index]

    def __call__(self, sample: Sample) -> dy.Expression:
        return self.build_sample(sample)

    def state_dict(self) -> Dict:
        weights = {}

        for meta, weight in zip(self.weights_meta, self.weights):
            if not meta.fixed and meta.index >= 0:
                weights[meta.index] = weight if isinstance(weight, (int, float)) else weight.value()
        return {"weights": weights}

    def load_state_dict(self, state_dict: Dict):
        weight_dict = state_dict["weights"]

        for i, (meta, weight) in enumerate(zip(self.weights_meta, self.weights)):
            if meta.index < 0 or meta.fixed:
                continue

            if isinstance(weight, (int, float)):
                self.weights[i] = weight_dict[meta.index]
            else:
                weight.set_value(np.array(weight_dict[meta.index]).reshape(weight.shape()))

    @staticmethod
    def to_dynet_value(value) -> dy.Expression:
        dim = 1
        if hasattr(value, "__len__"):
            dim = len(value)
        if dim == 1:
            return dy.scalarInput(float(value))
        if dim > 1:
            res = dy.vecInput(dim)
            res.set(value)
            return res
        return dy.inputTensor(value, dim)

    def to_dynet_expression(self, neuron: Neuron, neurons: List[dy.Expression], weights: List):
        if neuron.inputs:
            out = self.process_neuron_inputs(neuron, neurons, weights)
        else:
            out = [NeuraLogic.to_dynet_value(neuron.value)]

        if neuron.activation:
            if not neuron.pooling:
                out = sum(out)
            if neuron.activation != "Identity":
                out = NeuraLogic.activations[neuron.activation](out)
        else:
            out = sum(out)

        if self.hooks_set and neuron.hook_name is not None and neuron.hook_name in self.hooks:
            self.run_hook(neuron.hook_name, out.value())
        return out

    def process_neuron_inputs(self, neuron: Neuron, neurons: List, weights: List[dy.Parameters]) -> List[dy.Expression]:
        out = []

        if neuron.weights:
            for w, i in zip(neuron.weights, neuron.inputs):
                weight = weights[w]
                if self.weights_meta[w].fixed:
                    weight = dy.const_parameter(weight)
                if not isinstance(weight, int) and (neurons[i].dim()[0] == (1,) or weight.dim()[0] == (1,)):
                    out.append(dy.cmult(weight, neurons[i]))
                else:
                    out.append(weight * neurons[i])
        else:
            for i in neuron.inputs:
                out.append(neurons[i])
        return out



from typing import Optional, Dict, Union

import jpype

from neuralogic.core.enums import Backend
from neuralogic.core import Template, BuiltDataset
from neuralogic.nn.base import AbstractEvaluator
from neuralogic.core.settings import Settings

from neuralogic.dataset.base import BaseDataset


class JavaEvaluator(AbstractEvaluator):
    def __init__(
        self,
        problem: Optional[Template],
        settings: Settings,
    ):
        super().__init__(Backend.JAVA, problem, settings)

    def set_dataset(self, dataset: Union[BaseDataset, BuiltDataset]):
        super().set_dataset(dataset)
        self.neuralogic_model.set_training_samples(
            jpype.java.util.ArrayList([sample.java_sample for sample in self.dataset.samples])
        )

    def reset_dataset(self, dataset):
        if dataset is None:
            self.neuralogic_model.set_training_samples(jpype.java.util.ArrayList([]))
        else:
            self.neuralogic_model.set_training_samples(
                jpype.java.util.ArrayList([sample.java_sample for sample in dataset.samples])
            )
        self.dataset = dataset

    def train(
        self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True, epochs: int = None
    ):
        old_dataset = None

        if dataset is not None:
            old_dataset = self.dataset
            self.set_dataset(dataset)

        if epochs is None:
            epochs = self.settings.epochs

        def _train():
            for _ in range(epochs):
                results, total_len = self.neuralogic_model(None, True)
                yield sum(result[2] for result in results), total_len
            if dataset is not None:
                self.reset_dataset(old_dataset)

        if generator:
            return _train()

        results, total_len = self.neuralogic_model(None, True, epochs=epochs)
        if dataset is not None:
            self.reset_dataset(old_dataset)

        return sum(result[2] for result in results), total_len

    def test(self, dataset: Optional[Union[BaseDataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        def _test():
            for sample in dataset.samples:
                yield self.neuralogic_model(sample, False)

        if generator:
            return _test()
        return self.neuralogic_model(dataset.samples, False)

    def state_dict(self) -> Dict:
        return self.neuralogic_model.state_dict()

    def load_state_dict(self, state_dict: Dict):
        self.neuralogic_model.load_state_dict(state_dict)



from neuralogic.core.constructs.factories import Var, Constant, Relation, V, C, R
from neuralogic.core.template import Template
from neuralogic.core.builder import BuiltDataset
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.settings import Settings, SettingsProxy
from neuralogic.core.enums import Backend
from neuralogic.core.constructs.function import Transformation, Aggregation, Combination


__all__ = [
    "Var",
    "V",
    "Constant",
    "C",
    "Relation",
    "R",
    "Template",
    "BuiltDataset",
    "Transformation",
    "Aggregation",
    "Combination",
    "Backend",
    "Settings",
    "SettingsProxy",
    "Metadata",
]



class BaseDataset:
    def dump(
        self,
        queries_fp,
        examples_fp,
        sep: str = "\n",
    ):
        raise NotImplementedError

    def dump_to_file(
        self,
        queries_filename: str,
        examples_filename: str,
        sep: str = "\n",
    ):
        with open(queries_filename, "w") as queries_fp, open(examples_filename, "w") as examples_fp:
            self.dump(queries_fp, examples_fp, sep)


class ConvertableDataset(BaseDataset):
    def to_dataset(self):
        raise NotImplementedError
